%{
/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <stream.h>
#include <string.h>
#include <fcntl.h>

#include <unistd.h>

#include "yoo_root.hxx"
#include "dictionary.hxx"

struct node {
	char name[255];
	struct node *next;
};

void genVol() ;
int existServ(char* name) ;
int existVol(char* name) ;
int existClu(char* name) ;

extern char command[] ;
extern FILE *yyin ;

extern "C" {
  int yywrap() ;
  int yyparse() ;
}

#if defined(__alpha__) || (defined(__sun__) && !defined(__svr4__))
int yylex() ;
#endif

#if defined(__alpha__) || (defined(__sun__) && !defined(__svr4__))
int yyoutput(int c) ; 
#else
extern "C" { int yyoutput(int c) ; }
#endif



typedef Iterator<ADM_NameObj> __IterNameObj ;
typedef Iterator<ADM_Clu> __IterClu ;
typedef Iterator<ADM_Class> __IterClass ;

%}

%union  {
	char text[256];
	struct node *lst;
        }

%token <text> CREATE DELETE MOVE PRINT IN File ABORT COMMIT
%token <text> SCHEMA VOLUME VOLUMES CLUSTER CLUSTERS CLASS CLASSES
%token <text> IDENT FILENAME INT ID WITH DATA NAME NAMES QUIT
%token <text> SERVER SERVERS FOR LOG
%type <lst> identlst

%start cmdlst

%%
cmdlst:
	/* empty */
	| cmdlst cmd
	;

cmd:	volume
        | server
	| cluster 
	| class 
        | nameObj  
        | exit
	| error 
	;

server: createServer
        | deleteServer
	| setServLog
        | printServer

volume: createVol 
	| deleteVol 
	| moveVol 
	| printVol ;

cluster:createCluster 
	| deleteCluster 
	| printCluster ; 

class:	moveClass
	| printClass ; 

nameObj: printName
        | deleteName ;

printName: PRINT NAMES
	 {
	   cerr << "* exec : print names" << endl ;
	   Iterator<ADM_NameObj> iter(YoodaDictionary->dictName) ;
	   while (iter.more()) {
	     iter.get()->print() ;
	     iter.next() ;
	   }
	   cerr << "* done" << endl ;
	 }
	| PRINT NAMES identlst
	 {
	   cerr << "* exec : print names" << endl ;
	   for(node *cur = $3; cur; cur = cur->next) {
	     ADM_NameObj* nm = YoodaDictionary->dictName.getFirst(cur->name) ;
	     if (nm == NULL) {
	       cerr << "* Unknown name : " << cur->name << endl ;
	       continue ;
	     } else {
	       nm->print() ;
	     }
	   }
	   cerr << "* done" << endl ;
	 }
	;

deleteName :
	DELETE NAME IDENT
	 {
	   cerr << "* exec : delete name " << $3 << endl ;
	   ADM_NameObj* nm = YoodaDictionary->dictName.getFirst($3) ;
	   if (nm == NULL) {
	     cerr << "* Unknown name : " << $3 << endl ;
	   } else {
	     if (confirm("delete name")) {
	       delete nm ;
	       cerr << "* done" << endl ;
	     } else {
	       cerr << "* not deleted" << endl ;
	     }
	   }
	 }
        | DELETE NAMES
         {
	   cerr << "* exec : delete names" << endl ;
	   if (confirm("delete all names")) {
	     Iterator<ADM_NameObj> iter(YoodaDictionary->dictName) ;
	     while (iter.more()) {
	       delete (ADM_NameObj*)iter.get() ;
	       iter.next() ;
	     }
	     cerr << "* done" << endl ;
	   } else {
	     cerr << "* not deleted" << endl ;
	   }
	 }
	;

createServer:
	CREATE SERVER IDENT
	 {
	   if (existServ($3)) {
	     cerr << "* Server " << $3 << " already exists" << endl ;
	   } else {
	     cerr << "* exec : create server " << $3 ;
	     new ADM_Server($3) ;
	     cerr << "* done" << endl ;
	   }
	 }
	;

printServer:
        PRINT SERVERS
         {
	   cerr << "* exec : print servers" << endl ;
	   Iterator<ADM_Server> iter(YoodaDictionary->dictServ) ;
	   while (iter.more()) {
	     iter.get()->print(1) ;
	     iter.next() ;
	   }
	   cerr << "* done" << endl ;
	 }
	| PRINT SERVER IDENT
	 {
	   ADM_Server* server = YoodaDictionary->dictServ.getFirst($3) ;
           if (server == NULL) {
	     cerr << "* Unknown server : " << $3 << endl ;
	   } else {
	     server->print(1) ;
	   }
	 }
	;

setServLog:
	MOVE LOG FOR SERVER IDENT IN FILENAME
	 {
	   ADM_Server* server = YoodaDictionary->dictServ.getFirst($5) ;
           if (server == NULL) {
	     cerr << "* Unknown server : " << $3 << endl ;
	   } else {
	     server->movelog($7) ;
	   }
	 }
	;

deleteServer:
	DELETE SERVER IDENT
         {
	   cerr << "* exec : delete server " << $3 << endl ;
	   ADM_Server* server = YoodaDictionary->dictServ.getFirst($3) ;
           if (server == NULL) {
	     cerr << "* Unknown server : " << $3 << endl ;
	   } else {
	     if (server == YoodaDictionary->defServ) {
	       cerr << "* Impossible to delete default server" << endl ;
	     } else {		 
	       if (confirm("delete server")) {
		 delete server ;
		 cerr << "* done" << endl ;
	       } else {
		 cerr << "* not deleted" << endl ;
	       }
	     }
	   }
         }
	;

	
createVol:
	CREATE VOLUME IDENT
	 {
	   if (existVol($3)) {
	     cerr << "* volume " << $3 << " already exists" << endl ;
	   } else {
	     cerr << "* exec : create volume " << $3 << endl ;
	     new ADM_Vol($3,NULL) ;
	     cerr << "* done" << endl ;
	   }
	 }
	| CREATE VOLUME IDENT IN SERVER IDENT
	 {
	   if (existVol($3)) {
	     cerr << "* volume " << $3 << " already exists" << endl ;
	   } else {
	     ADM_Server* server = YoodaDictionary->dictServ.getFirst($6) ;
	     if (server == NULL) {
	       cerr << "* Unknown server : " << $3 << endl ;
	     } else {
	       cerr << "* exec : create volume " << $3 << endl ;
	       new ADM_Vol($3,NULL,server) ;
	       cerr << "* done" << endl ;
	     }
	   }
	 }
	| CREATE VOLUME IDENT IN File FILENAME
         {
	   if (existVol($3)) {
	     cerr << "* volume " << $3 << " already exists" << endl ;
	   } else {
	     cerr << "* exec : create volume " << $3 ;
	     cerr << " in file " << $6 << endl ;
	     new ADM_Vol($3,$6) ;
	     cerr << "* done" << endl ;
	   }
	 }
	| CREATE VOLUME IDENT IN File FILENAME IN SERVER IDENT
         {
	   if (existVol($3)) {
	     cerr << "* volume " << $3 << " already exists" << endl ;
	   } else {
	     ADM_Server* server = YoodaDictionary->dictServ.getFirst($9) ;
	     if (server == NULL) {
	       cerr << "* Unknown server : " << $3 << endl ;
	     } else {
	       cerr << "* exec : create volume " << $3 ;
	       cerr << " in file " << $6 << endl ;
	       new ADM_Vol($3,$6,server) ;
	       cerr << "* done" << endl ;
	     }
	   }
	 }
	;


deleteVol: 
	DELETE VOLUME IDENT 
	 {
	   cerr << "* exec : delete volume " << $3 << endl ;
	   ADM_Vol* volume = YoodaDictionary->dictVol.getFirst($3) ;
	   if (volume == YoodaDictionary->defVol) {
	     cerr << "* Impossible to delete default volume" << endl ;
	   } else {
	     if (volume == NULL) {
	       cerr << "* Unknown volume : " << $3 << endl ;
	     } else {
	       if (confirm("delete volume")) {
		 delete volume ;
		 cerr << "* done" << endl ;
	       } else {
		 cerr << "* not deleted" << endl ;
	       }
	     }
	   }
 	 }
        | DELETE VOLUMES
         {
	   cerr << "* exec : delete volumes" << endl ;
	   if (confirm("delete all volumes")) {
	     Iterator<ADM_Vol> iter(YoodaDictionary->dictVol) ;
	     while (iter.more()) {
	       if (iter.get() != YoodaDictionary->defVol) {
		 delete (ADM_Vol*)iter.get() ;
		 iter.next() ;
	       }
	     }
	     cerr << "* done" << endl ;
	   } else {
	     cerr << "* not deleted" << endl ;
	   }
	 }
	;

moveVol: 
	MOVE VOLUME IDENT IN File FILENAME
         {
	   cerr << "* exec : move volume " << $3 << " in file " << $6 << endl ;
	   ADM_Vol* volume = YoodaDictionary->dictVol.getFirst($3) ;
	   if (volume == NULL) {
	     cerr << "* Unknown volume : " << $3 << endl ;
	   } else {
	     volume->movefile($6) ;
	     cerr << "* done" << endl ;
	   }
	 }
	| MOVE VOLUME IDENT IN SERVER IDENT
         {
	  cerr << "* exec : move volume " << $3 << " in server " << $6 << endl;
	   ADM_Vol* volume = YoodaDictionary->dictVol.getFirst($3) ;
	   if (volume == NULL) {
	     cerr << "* Unknown volume : " << $3 << endl ;
	   } else {
	     ADM_Server* server = YoodaDictionary->dictServ.getFirst($6) ;
	     if (server == NULL) {
	       cerr << "* Unknown server : " << $3 << endl ;
	     } else {
	       volume->moveserver(server) ;
	       cerr << "* done" << endl ;
	     }
	   }
	 }
	;

printVol:
	PRINT VOLUMES
	 {
	   cerr << "* exec : print volumes" << endl ;
	   Iterator<ADM_Vol> iter(YoodaDictionary->dictVol) ;
	   while (iter.more()) {
	     iter.get()->print(1) ;
	     iter.next() ;
	   }
	   cerr << "* done" << endl ;
	 }
	| PRINT VOLUME identlst
	 {
	   cerr << "* exec : print volumes" << endl ;
	   for(node *cur = $3; cur; cur = cur->next) {
	     ADM_Vol* volume = YoodaDictionary->dictVol.getFirst(cur->name) ;
	     if (volume == NULL) {
	       cerr << "* Unknown volume : " << cur->name << endl ;
	       continue ;
	     } else {
	       volume->print(1) ;
	     }
	   }
	   cerr << "* done" << endl ;
	 }
	;

identlst:
	IDENT
	  {
	    $$ = new node;
	    strcpy($$->name, $1) ;
	    $$->next = NULL ;
	  }
	| IDENT identlst
	  {
	    node *ident = new node;
	    strcpy(ident->name, $1);
	    ident->next = $2;
	    $$ = ident;
	  }
	;

createCluster:
	CREATE CLUSTER IDENT
	 {
	   if (existClu($3)) {
	     cerr << "* cluster " << $3 << " already exists" << endl ;
	   } else {
	     cerr << "* exec : create cluster " << $3 ;
	     cerr << " in default volume" << endl ;
	     new ADM_Clu($3, YoodaDictionary->defVol) ;
	     cerr << "* done" << endl ;
	   }
	 }
	| CREATE CLUSTER IDENT IN VOLUME IDENT
         {
	   if (existClu($3)) {
	     cerr << "* cluster " << $3 << " already exists" << endl ;
	   } else {
	     cerr << "* exec : create cluster " << $3 ;
	     cerr << " in volume " << $6 << endl ;
	     ADM_Vol* volume = YoodaDictionary->dictVol.getFirst($6) ;
	     if (volume == NULL) {
	       cerr << "* Unknown volume : " << $6 << endl ;
	     } else {
	       new ADM_Clu($3,volume) ;
	       cerr << "* done" << endl ;
	     }
	   }
	 }
	;

deleteCluster:
	DELETE CLUSTER IDENT
         {
	   cerr << "* exec : delete cluster " << $3 << endl ;
	   ADM_Clu* cluster = YoodaDictionary->dictClu.getFirst($3) ;
	   if (cluster == YoodaDictionary->defClu) {
	     cerr << "* Impossible to delete default cluster" << endl ;	     
	   } else {
	     if (cluster == NULL) {
	       cerr << "* Unknown cluster : " << $3 << endl ;
	     } else {
	       if (confirm("delete cluster")) {
		 delete cluster ;
		 cerr << "* done" << endl ;
	       } else {
		 cerr << "* not deleted" << endl ;
	       }
	     }
	   }
 	 }
        | DELETE CLUSTERS
         {
	   cerr << "* exec : delete clusters" << endl ;
	   if (confirm("delete all clusters")) {
	     Iterator<ADM_Clu> iter(YoodaDictionary->dictClu) ;
	     while (iter.more()) {
	       if (iter.get() != YoodaDictionary->defClu) {
		 delete (ADM_Clu*)iter.get() ;
		 iter.next() ;
	       }
	     }
	     cerr << "* done" << endl ;
	   } else {
	     cerr << "* not deleted" << endl ;
	   }
	 }
	;

printCluster:
	PRINT CLUSTERS
	 {
	   cerr << "* exec : print clusters" << endl ;
	   Iterator<ADM_Clu> iter(YoodaDictionary->dictClu) ;
	   while (iter.more()) {
	     iter.get()->print(1) ;
	     iter.next() ;
	   }
	   cerr << "* done" << endl ;
	 }
	| PRINT CLUSTER identlst
	 {
	   cerr << "* exec : print clusters" << endl ;
	   for(node *cur = $3; cur; cur = cur->next) {
	     ADM_Clu* clu = YoodaDictionary->dictClu.getFirst(cur->name) ;
	     if (clu == NULL) {
	       cerr << "* Unknown cluster : " << cur->name << endl ;
	       continue ;
	     } else {
	       clu->print(1) ;
	     }
	   }
	   cerr << "* done" << endl ;
	 }
	;


moveClass:
	MOVE CLASS IDENT IN CLUSTER IDENT
        {
	   cerr << "* exec : move class " << $3 << " in cluster " ;
	   cerr << $6 << endl ;
	   ADM_Class* cl = YoodaDictionary->dictClass.getFirst($3) ;
	   ADM_Clu* clu = YoodaDictionary->dictClu.getFirst($6) ;
	   if (cl == NULL) {
	     cerr << "* Unknown class : " << $3 << endl ;
	   } else if (clu == NULL) {
	     cerr << "* Unknown cluster : " << $6 << endl ;
	   } else {
	     cl->move(clu) ;
	     cerr << "* done" << endl ;
	   }
	 }
	;

printClass:
	PRINT CLASSES
	 {
	   cerr << "* exec : print classes" << endl ;
	   Iterator<ADM_Class> iter(YoodaDictionary->dictClass) ;
	   while (iter.more()) {
	     iter.get()->print() ;
	     iter.next() ;
	   }
	   cerr << "* done" << endl ;
	 }
	| PRINT CLASS identlst
	 {
	   cerr << "* exec : print classes" << endl ;
	   for(node *cur = $3; cur; cur = cur->next) {
	     ADM_Class* cl = YoodaDictionary->dictClass.getFirst(cur->name) ;
	     if (cl == NULL) {
	       cerr << "* Unknown class : " << cur->name << endl ;
	       continue ;
	     } else {
	       cl->print() ;
	     }
	   }
	   cerr << "* done" << endl ;
	 }
	;

exit:  ABORT
       {
	 command[0] = '\0' ;
	 yoo_abort() ;
       } 
       | COMMIT
       {
	 genVol() ;
	 system(command) ;
	 yoo_commit() ;
       }
       | QUIT
       {
	 return 1;
       }

%%

int existServ(char* name) {

  return (YoodaDictionary->dictServ.getFirst(name) != NULL) ;
}

int existVol(char* name) {

  return (YoodaDictionary->dictVol.getFirst(name) != NULL) ;
}

int existClu(char* name) {

  return (YoodaDictionary->dictClu.getFirst(name) != NULL) ;
}

void shell() {

  command[0] = 0 ;
  char YOODA_tmp[256] ;
  FILE *f_ref;
  char c;
  printf("enter command\n");
  f_ref = fopen(tmpnam(YOODA_tmp), "w") ;
  while(scanf("%c", &c) != EOF) {
    if(c != '\n') { fprintf(f_ref, "%c", c); continue ; }
    fclose(f_ref);
    yyin = fopen(YOODA_tmp, "r");
    int res = yyparse() ; 
    fclose(yyin);
    unlink(YOODA_tmp);
    if (res == 1) break ;
    printf("enter command\n");
    f_ref = fopen(tmpnam(YOODA_tmp), "w") ;
  }
}

extern "C" {
void yyerror(char *s) {
  printf("admin %s\n", s);
}
}

int confirm(char* s) {

  char res[500] ;

  do {
    cerr << "*\tConfirm : " << s << " ? [y/n] " ;
    cin >> res ;
  } while ( (strcmp(res,"y")!=0) && (strcmp(res, "n")!=0) ) ;

  if (strcmp(res, "y") == 0) {
    return 1 ;
  } else {
    return 0 ;
  }
}
